// -*- C++ -*-
/*!
 * @file PeriodicTask.h
 * @brief PeriodicTask class
 * @date $Date$
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2009
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id$
 *
 */

#ifndef COIL_PERIODICTASK_H
#define COIL_PERIODICTASK_H

#include <map>
#include <string>
#include <vector>
#include <algorithm>

#include <coil/Mutex.h>
#include <coil/Guard.h>
#include <coil/Condition.h>
#include <coil/TimeValue.h>
#include <coil/TimeMeasure.h>
#include <coil/PeriodicTaskBase.h>

namespace coil
{
  /*!
   * @if jp
   * @class PeriodicTask
   * @brief ^XNXbhsNX
   *
   * ̊֐s邽߂̃XbhIuWFNgB
   * gp菇͈ȉ̒ʂB
   *
   * task; // CX^X
   * task.setTask(TaskFuncBase(obj, mem_func)); // s֐^
   * task.activate(); // XbhX^[g
   *
   * task.suspend(); // s~߂
   * task.signal(); // 1s
   * task.resume(); // sĊJ
   *
   * task.finalize(); // ^XNI
   * 
   * @else
   * @class PeriodicTask
   * @brief PeriodicTask class
   *
   * @endif
   */
  class PeriodicTask
    : public coil::PeriodicTaskBase
  {
  public:
    typedef coil::Guard<coil::Mutex> Guard;

    /*!
     * @if jp
     *
     * @brief RXgN^
     *
     * RXgN^
     *
     * @else
     *
     * @brief Constructor
     *
     * Constructor
     *
     * @endif
     */
    PeriodicTask();
    
    /*!
     * @if jp
     *
     * @brief fXgN^
     *
     * fXgN^B
     *
     * @else
     *
     * @brief Destructor
     *
     * Destructor
     *
     * @endif
     */
    virtual ~PeriodicTask();
    
    /*!
     * @if jp
     * @brief ^XNsJn
     *
     * ^XN̎sJn邽߂ɃXbhX^[gB  ^XN
     * ɊJnꂽꍇ true ԂAłɃ^XNJnς݁A܂
     * ͎s^XNݒ肳ĂȂ false ԂB
     *
     * @return true: JnAfalse: Xbhς݂^XNݒłB
     *
     * @else
     * @brief Starting the task
     *
     * Starting a thread to execute a task.  If the task/thread is
     * started properly, it will return 'TRUE'.  if the task/thread
     * are already started or task function object is not set, 'FALSE'
     * will be returned.
     *
     * @return true: normal start, false: already started  or task is not set
     *
     * @endif
     */
    virtual void activate();

    /*!
     * @if jp
     * @brief ^XNsI
     *
     * s̃^XNIB
     *
     * @else
     * @brief Finalizing the task
     *
     * Finalizing the task running.
     *
     * @endif
     */
    virtual void finalize();

    /*!
     * @if jp
     * @brief ^XNs𒆒f
     *
     * s̃^XN𒆒fB
     *
     * @else
     * @brief Suspending the task
     *
     * Suspending the task running.
     *
     * @endif
     */
    virtual int suspend(void);

    /*!
     * @if jp
     * @brief fĂ^XNĊJ
     *
     * fĂ^XNĊJ
     *
     * @else
     * @brief Resuming the suspended task
     *
     * Resuming the suspended task
     *
     * @endif
     */
    virtual int resume(void);

    /*!
     * @if jp
     * @brief fĂ^XN1s
     *
     * fĂ^XN1s
     *
     * @else
     * @brief Executing the suspended task one tick
     *
     * Executing the suspended task one tick
     *
     * @endif
     */
    virtual void signal();

    /*!
     * @if jp
     * @brief ^XNs֐Zbg
     *
     * @param func int (*)() ^̊֐|C^
     *
     * @else
     * @brief Setting task execution function
     *
     * @param func Set int (*)() type function pointer
     *
     * @endif
     */
    virtual bool setTask(TaskFuncBase* func, bool delete_in_dtor = true);

    /*!
     * @if jp
     * @brief ^XNs֐Zbg
     *
     * @param func int (*)() ^̊֐|C^
     *
     * @return true: , false: s
     *
     * @else
     * @brief Setting task execution function
     *
     * @param func Set int (*)() type function pointer
     *
     * @return true: successful, false: failed
     *
     * @endif
     */
    template <class O, class F>
    bool setTask(O* obj, F fun)
    {
      return this->setTask(new TaskFunc<O, F>(obj, fun));
    }

    /*!
     * @if jp
     * @brief ^XNsZbg
     *
     * @param period s [sec]
     *
     * @else
     * @brief Setting task execution period
     *
     * @param period Execution period [sec]
     *
     * @endif
     */
    virtual void setPeriod(double period);

    /*!
     * @if jp
     * @brief ^XNsZbg
     *
     * @param period s
     *
     * @else
     * @brief Setting task execution period
     *
     * @param period Execution period
     *
     * @endif
     */
    virtual void setPeriod(TimeValue& period);

    /*!
     * @if jp
     * @brief ^XN֐sԌvLɂ邩
     * @else
     * @brief Validate a Task execute time measurement
     * @endif
     */
    virtual void executionMeasure(bool value);
    
    /*!
     * @if jp
     * @brief ^XN֐sԌv
     * @else
     * @brief Task execute time measurement period
     * @endif
     */
    virtual void executionMeasureCount(int n);
    
    /*!
     * @if jp
     * @brief ^XNԌvLɂ邩
     * @else
     * @brief Validate a Task period time measurement
     * @endif
     */
    virtual void periodicMeasure(bool value);
    
    /*!
     * @if jp
     * @brief ^XNԌv
     * @else
     * @brief Task period time measurement count
     * @endif
     */
    virtual void periodicMeasureCount(int n);
    
    /*!
     * @if jp
     * @brief ^XN֐sԌvʂ擾
     * @else
     * @brief Get a result in task execute time measurement
     * @endif
     */
    virtual TimeMeasure::Statistics getExecStat();
    
    /*!
     * @if jp
     * @brief ^XNԌvʂ擾
     * @else
     * @brief Get a result in task period time measurement
     * @endif
     */
    virtual TimeMeasure::Statistics getPeriodStat();
    
  protected:
    /*!
     * @if jp
     * @brief PeriodicTask p̃Xbhs
     * @else
     * @brief Thread execution for PeriodicTask
     * @endif
     */
    virtual int svc();

    /*!
     * @if jp
     * @brief Xbhx~
     * @else
     * @brief Thread sleep
     * @endif
     */
    virtual void sleep();

    /*!
     * @if jp
     * @brief sԍXV
     * @else
     * @brief Update for execute state
     * @endif
     */
    virtual void updateExecStat();

    /*!
     * @if jp
     * @brief ԍXV
     * @else
     * @brief Update for period state
     * @endif
     */
    virtual void updatePeriodStat();

  protected:
    /*!
     * @if jp
     * @brief ^XNs
     * @else
     * @brief Task execution period
     * @endif
     */
    coil::TimeValue m_period;

    /*!
     * @if jp
     * @brief Xbhx~tO
     * @else
     * @brief Thread sleep flag
     * @endif
     */
    bool m_nowait;

    /*!
     * @if jp
     * @brief ^XNs֐
     * @else
     * @brief Task execution function
     * @endif
     */
    TaskFuncBase* m_func;

    /*!
     * @if jp
     * @brief ^XNs֐폜tO
     * @else
     * @brief Task execution function delete flag
     * @endif
     */
    bool m_deleteInDtor;

    /*!
     * @if jp
     * @class alive_t
     * @brief alive_t NX
     * @else
     * @class alive_t
     * @brief alive_t class
     * @endif
     */
    class alive_t
    {
    public:
      alive_t(bool val) : value(val) {}
      bool value;
      coil::Mutex mutex;
    };

    /*!
     * @if jp
     * @brief ^XNtO
     * @else
     * @brief Task alive flag
     * @endif
     */
    alive_t m_alive;

    /*!
     * @if jp
     * @brief ^XNfǗp\
     * @else
     * @brief Structure for task suspend management
     * @endif
     */
    struct suspend_t
    {
      suspend_t(bool sus) : suspend(sus), mutex(), cond(mutex) {}
      bool suspend;
      coil::Mutex mutex;
      coil::Condition<coil::Mutex> cond;
    };

    /*!
     * @if jp
     * @brief ^XNf
     * @else
     * @brief Task suspend infomation
     * @endif
     */
    suspend_t m_suspend;
      
    /*!
     * @if jp
     * @brief ^XNsԌvǗp\
     * @else
     * @brief Structure for task execution time measurement management
     * @endif
     */
    struct statistics_t
    {
      coil::TimeMeasure::Statistics stat;
      coil::Mutex mutex;
    };

    /*!
     * @if jp
     * @brief ^XNsԌvtO
     * @else
     * @brief Task execution time measurement flag
     * @endif
     */
    bool              m_execMeasure;

    /*!
     * @if jp
     * @brief ^XNsԌv
     * @else
     * @brief Task execution time measurement count
     * @endif
     */
    unsigned int      m_execCount;

    /*!
     * @if jp
     * @brief ^XNsԌv
     * @else
     * @brief Task execution time measurement max count
     * @endif
     */
    unsigned int      m_execCountMax;

    /*!
     * @if jp
     * @brief ^XNsԌvv
     * @else
     * @brief Task execution time measurement statistics
     * @endif
     */
    statistics_t      m_execStat;

    /*!
     * @if jp
     * @brief ^XNsԌv
     * @else
     * @brief Task execution time  measurement infomation
     * @endif
     */
    coil::TimeMeasure m_execTime;

    /*!
     * @if jp
     * @brief ^XNԌvtO
     * @else
     * @brief Task periodic time measurement flag
     * @endif
     */
    bool              m_periodMeasure;

    /*!
     * @if jp
     * @brief ^XNԌv
     * @else
     * @brief Task periodic time measurement count
     * @endif
     */
    unsigned int      m_periodCount;

    /*!
     * @if jp
     * @brief ^XNԌvő吔
     * @else
     * @brief Task periodic time measurement max count
     * @endif
     */
    unsigned int      m_periodCountMax;

    /*!
     * @if jp
     * @brief ^XNԌvv
     * @else
     * @brief Task periodic time measurement statistics
     * @endif
     */
    statistics_t      m_periodStat;

    /*!
     * @if jp
     * @brief ^XNԌv
     * @else
     * @brief Task periodic time  measurement infomation
     * @endif
     */
    coil::TimeMeasure m_periodTime;

  };

}; // namespace coil

#endif // COIL_PERIODICTASK_H
